"""
This script will migrate a privileged phantom installation to unprivileged. It
s shipped with the privileged installer and will do the migration in-place.
"""
import argparse
from collections import namedtuple
from configparser import RawConfigParser
from datetime import datetime
import shutil
import sys
import re
import pwd
from time import sleep
from pathlib import Path
import json
import os
import logging
import subprocess
import socket
from typing import List, Optional

from phantom_common.phlog import init_logging
from phantom_common.paths import (
    PHANTOM_BIN,
    PHANTOM_HOME,
    PHANTOM_USER_HOME,
    PHANTOM_INSTALL_MANIFEST,
)
from phantom_common.cluster import cluster_constants
from phantom_common.install_info import (
    get_phantom_user,
    get_phantom_group,
    get_product_name,
    PhantomInstallConf,
)
from phantom_common.shellpy import response

from manifest import InstallManifest

logger = logging.getLogger(__name__)
PHANTOM_USER = get_phantom_user()
PHANTOM_GROUP = get_phantom_group()
PRODUCT_NAME = get_product_name()
CRON_HEADER = "### START OF PHANTOM JOBS - KEEP THEM AS THEY ARE ###"
CRON_FOOTER = "### END OF PHANTOM JOBS #############################"


class UserDeclinedToContinue(RuntimeError):
    pass


class PhCalledProcessException(Exception):
    def __init__(self, message, proc):
        self.proc = proc
        super().__init__(message)


def prompt_for_confirmation(question, default=True):
    if default:
        default_text = "[Y/n]"
    else:
        default_text = "[y/N]"

    answer = input(f"{question} {default_text} ")

    if not any(
        [response.is_empty(answer), response.is_affirmative(answer), response.is_negative(answer)]
    ):
        logger.info(f'"{answer}" is not a valid answer.')
        return prompt_for_confirmation(question, default)

    if not response.is_affirmative(answer):
        raise UserDeclinedToContinue


# Packages that should be uninstalled the regular way
RPM_UNINSTALL = (
    "pgbouncer",
    "nginx",
    "postgresql11-server",
    "git",
)

MIGRATION_DESCRIPTION = f"""As part of the migration, the following will take place (non-exhaustive list):

  - {PRODUCT_NAME} RPMs will be removed from the RPM database on this system
  - External dependencies (Postgres, pgbouncer, nginx, git) will be uninstalled
    and replaced by unprivileged versions living under {PHANTOM_HOME}
  - Ownership of all postgres databases and tables will be transferred to 'phantom' role
    and all other roles will be removed.
  - {PRODUCT_NAME} logging will be moved to {PHANTOM_HOME}/var/log/phantom
  - {PRODUCT_NAME} Rsyslog configuration will be removed.
  - Replace the contents of /home/phantom/.ssh

THIS MIGRATION IS NOT REVERSIBLE!
"""

ExtFileMigrationEntry = namedtuple("ExtFileMigrationEntry", "source target operation")


class MigratePrivilegedInstallToUnprivileged:
    """Tool to migrate a privileged install to unprivileged."""

    _CLUSTER_PHASE_FIRST_HALF = "FIRST_HALF"

    EXT_FILES_TO_MIGRATE = (
        # Move these, since we may need some of the contents.
        ExtFileMigrationEntry(
            Path("/etc/pgbouncer/pgbouncer.database.ini"),
            "etc/pgbouncer/pgbouncer.database.ini",
            "move",
        ),
        ExtFileMigrationEntry(
            Path("/etc/pgbouncer/userlist.txt"), "etc/pgbouncer/userlist.txt", "move"
        ),
        ExtFileMigrationEntry(Path("/var/log/phantom"), "var/log/phantom", "move"),
        # Copy, rather than move certain logs that might be used for non-phantom things
        ExtFileMigrationEntry(Path("/var/log/nginx"), "var/log/nginx", "copy"),
        ExtFileMigrationEntry(Path("/var/log/pgbouncer"), "var/log/pgbouncer", "copy"),
        ExtFileMigrationEntry(Path("/var/log/pgbackrest"), "var/log/pgbackrest", "copy"),
    )

    def __init__(self, **kwargs):
        self.no_prompt = kwargs.get("no_prompt", False)
        self.cluster_phase = kwargs.get("cluster_phase", None)
        self.local_https_port = self.external_https_port = str(kwargs.get("https_port", "9999"))
        self.phantom_home = Path(PHANTOM_HOME)
        self.phantom_log = self.phantom_home / "var/log/phantom"
        self.phantom_etc = self.phantom_home / "etc"
        self.phantom_var = self.phantom_home / "var"
        self.phantom_usr = self.phantom_home / "usr"
        self._phsvc = self.phantom_home / "bin/phsvc"
        self._phenv = self.phantom_home / "bin/phenv"
        self.manifest = InstallManifest(False, Path(PHANTOM_INSTALL_MANIFEST))
        self.migration_component_dir = self.phantom_home / "migration"

        logger.debug("Investigating system.")
        self.external_db = self._has_external_db()

        # TODO: Verify clustering
        self.is_clustered = self._is_clustering_enabled()
        if self.is_clustered:
            self.external_https_port = "443"

    def apply(self):
        self._ensure_db_is_up()
        self.migrate()

    def migrate(self):
        if not self.no_prompt:
            logger.info(
                f"You are about to convert this {PRODUCT_NAME} installation to an unprivileged one."
                f"\n{MIGRATION_DESCRIPTION}"
            )
            prompt_for_confirmation("Do you want to proceed?", default=False)

        logger.info(
            f"Migrating a privileged {PRODUCT_NAME} installation to an unprivileged installation."
        )
        os.umask(0o22)

        # Disable systemd services
        self.stop_and_disable_services()

        # Stop daemons, so we can safely update daemon logging settings on clusters.
        self.stop_daemons()

        # Update daemon logging settings. Requires db to be up.
        self.configure_daemon_logging()

        if not self.external_db:
            # Postgres roles in the db need adjusting. Requires db to be up.
            self.configure_postgres_roles()

        # Stop the rest of phantom (postgres,nginx,splunk,etc), so we can safely copy/move files
        self.stop_phantom()

        # Copy/Move files
        self.migrate_files()

        # Remove rpm packages from db
        self.remove_rpm_packages_from_db()

        # Install the new dependencies
        self.install_files()

        # Update config files
        self.configure_nri()

        # Remove the packages that should be removed
        self.uninstall_rpm_packages()

        # Start phantom
        self.start_phantom()

        # Remove migration files
        self.remove_migration_component()

        logger.info("Migration complete.")

    def stop_daemons(self):
        logger.info(f"Stopping {PRODUCT_NAME} daemons.")
        self._run(f'{Path(PHANTOM_BIN) / "stop_daemon.sh"} all')

    def stop_phantom(self):
        """Stop the remaining services.

        Running stop_phantom.sh as root on an already migrated system will
        leave pgbouncer and postgres running, so we try to run it as phantom
        user first, and then run it again as root if that fails.
        """
        logger.info(f"Stopping remaining {PRODUCT_NAME} services.")
        stop_phantom_cmd = os.path.join(PHANTOM_BIN, "stop_phantom.sh")
        try:
            self._run(stop_phantom_cmd, run_as_phantom_user=True)
        except PhCalledProcessException as e:
            if "Error: must run with root privileges" in e.proc.stdout:
                self._run(stop_phantom_cmd)
            else:
                raise

    def stop_and_disable_services(self):
        logger.info("Disabling services.")

        services_to_disable = ("phantom_watchdogd", "uwsgi")
        for service in services_to_disable:
            try:
                self._run(f"{self._phsvc} disable {service}")
            except PhCalledProcessException as e:
                if f"supported commands for {service}" in e.proc.stdout:
                    logger.debug(f"{service} was already disabled.")
                else:
                    raise

        services_to_stop = ("phantom_watchdogd", "phantom_supervisord", "uwsgi")
        for service in services_to_stop:
            self._run(f"{self._phsvc} stop {service}")

    def start_phantom(self):
        """Start phantom again."""
        logger.info(f"Starting {PRODUCT_NAME}.")
        start_phantom_sh = os.path.join(PHANTOM_BIN, "start_phantom.sh")
        try:
            self._run(start_phantom_sh, timeout=120, run_as_phantom_user=True)
        except PhCalledProcessException:
            # Allow startup failures when upgrade migrating clusters, and
            # we've yet to reach the inflection point.
            if self.is_clustered and self.cluster_phase == self._CLUSTER_PHASE_FIRST_HALF:
                return

            raise

    def _has_external_db(self):
        pgbouncer_ini_locations = [
            Path("/etc/pgbouncer/pgbouncer.database.ini"),
            self.phantom_home / "etc/pgbouncer/pgbouncer.database.ini",
        ]
        for pgbouncer_ini in pgbouncer_ini_locations:
            if os.path.isfile(pgbouncer_ini):
                parser = RawConfigParser(allow_no_value=True)
                parser.read(pgbouncer_ini)
                try:
                    for line in parser["databases"].values():
                        return "host=" in line
                except KeyError:
                    continue
        else:
            logger.error(
                f"Failed to detect if this {PRODUCT_NAME} "
                f"installation is using an external database."
            )
            sys.exit(-1)

    def _ensure_db_is_up(self, retry_no=0):
        """Check that we can connect to the database and if not, attempt to fix that."""
        if retry_no >= 6:
            logger.error(
                "Failed to connect to the database. Please make sure that the database and pgbouncer are "
                "running and then run this command again."
            )
            exit(-1)
        try:
            return self._run_sql("select 0", run_as_phantom_user=True)

        except PhCalledProcessException as e:
            if "could not connect to server" in str(e):
                logger.debug("Pgbouncer is down. Trying to start it.")
                pgbouncer_bin = self.phantom_home / "bin/pgbouncer"
                pgbouncer_migrated = pgbouncer_bin.exists()

                try:
                    self._run(
                        f"{self._phsvc} start pgbouncer", run_as_phantom_user=pgbouncer_migrated
                    )
                except PhCalledProcessException as e:
                    logger.error(
                        "Pgbouncer is down and we have failed to start it. Please make sure it is running "
                        "and then run this command again"
                    )
                    exit(-1)

                sleep(10)

            elif "pgbouncer cannot connect to server" in str(e):
                if self.external_db:
                    logger.error(
                        "This instance seems to be using an external database. "
                        "Please make sure it is running and then run this command again."
                    )
                    exit(-1)
                postgresql_bin = self.phantom_home / "bin/postgresql"
                postgresql_migrated = postgresql_bin.exists()
                postgresql_is_running = (
                    self._run(
                        f"{self._phsvc} status postgresql-11",
                        check=False,
                        run_as_phantom_user=postgresql_migrated,
                    ).returncode
                    == 0
                )

                if not postgresql_is_running:
                    logger.debug("Database is down. Trying to start it.")

                    try:
                        self._run(
                            f"{self._phsvc} start postgresql-11",
                            run_as_phantom_user=postgresql_migrated,
                        )
                    except PhCalledProcessException:
                        logger.error(
                            "Postgresql is down and we have failed to start it. Please make sure it is running "
                            "and then run this command again"
                        )
                        exit(-1)

                    sleep(5)
                else:
                    logger.debug(
                        "Database is up, but not accepting connections yet. Sleeping a bit."
                    )
                    sleep(15)

        except subprocess.TimeoutExpired:
            pass

        retry_no += 1
        self._ensure_db_is_up(retry_no)

    def _modify_line_in_file(self, file_path, search, replacement):
        self._run(f'sed -i "s,{search},{replacement},g" {file_path}')

    def _is_clustering_enabled(self):
        cluster_check = self.phantom_home / "bin/is_clustering_enabled.pyc"
        proc = self._run(
            f"{self._phenv} python {cluster_check}", check=False, run_as_phantom_user=True
        )
        if proc.returncode == 0:
            return True
        elif proc.returncode == 1:
            return False
        else:
            raise Exception(
                f"Could not detect if this {PRODUCT_NAME} instance is part of a cluster."
            )

    def uninstall_rpm_packages(self):
        """Uninstall packages.

        Prompt for confirmation with each package, unless the appropriate command-line argument
        was given.
        """
        logger.info("Removing packages that will be replaced with their unprivileged equivalents.")
        for package in RPM_UNINSTALL:
            try:
                self._run(f"rpm -e {package}")
            except PhCalledProcessException as e:
                if "not installed" in str(e):
                    logger.debug(f"{package} has already been uninstalled.")
                else:
                    raise
            else:
                logger.debug(f"Uninstalled {package}")

    def remove_rpm_packages_from_db(self):
        """Remove packages from rpm db.

        The contents of these packages are the same on NRI and privileged install,
        so we just remove the packages from the rpm db, but don't remove the files.
        """
        logger.info(f"Removing {PRODUCT_NAME} packages from RPM database.")
        try:
            self._run("rpm -qa | grep phantom | sudo xargs rpm -e --justdb", timeout=None)
        except PhCalledProcessException as e:
            # If this command has already been run, we get an error
            # "rpm: no packages given for erase" which is safe to ignore.
            if "no packages given for erase" not in str(e):
                raise

    def migrate_files(self):
        """Copy/move/delete files and folders.

        Delete files that won't be needed or will be replaced from templates.
        Move files/folders that need to be modified or can be used as is.
        Copy files/folders that might contain non-phantom content.
        """
        entries_to_remove: List[Path] = []
        entries_to_add: List[Path] = []

        logger.info("Migrating files.")
        manifest = self.manifest.read()

        for source, destination, method in self.EXT_FILES_TO_MIGRATE:
            destination = self.phantom_home / destination

            # Make sure parent dir exists
            destination.parent.mkdir(parents=True, exist_ok=True)

            source_exists_in_manifest = source in manifest

            if source_exists_in_manifest:
                entries_to_remove.append(source)

            if method == "move":
                # Check if we've already moved the file
                if not source.exists() and destination.exists():
                    logger.debug(f"{source} has already been moved to {destination}")
                    continue

                if source_exists_in_manifest:
                    entries_to_add.append(destination)

                shutil.move(source, destination)

            elif method == "copy":
                # Copying files is safe to do even on subsequent runs of this script.
                if source.is_file():
                    shutil.copy(source, destination)
                else:
                    if destination.exists():
                        logger.debug("{source} has already been copied to {destination}")
                        continue
                    shutil.copytree(source, destination)

                if source_exists_in_manifest:
                    entries_to_add.append(destination)

            elif method == "delete":
                if not source.exists():
                    logger.debug(f"{source} has already been deleted.")
                    continue

                if source.is_file():
                    source.unlink()
                else:
                    shutil.rmtree(source)
            else:
                # This indicates a developer mistake, should never reach this code path
                raise ValueError(f'Invalid method "{method}" for file "{source}"')

            self.manifest.remove_entries(*entries_to_remove)
            self.manifest.remove_entries(*entries_to_add)

    def _change_file_ownerships(self):
        """Phantom-user should own everything inside phantom home."""
        logger.debug(
            f"Modifying file ownerships so that {PHANTOM_USER}-user owns everything in {PHANTOM_HOME}. "
            f"This may take a long time."
        )
        self._run(
            f"chown -R {PHANTOM_USER}:{PHANTOM_GROUP} {PHANTOM_HOME}",
            timeout=None,
        )

        manifest = self.manifest.read()
        for path_str, entry in manifest.items():
            path = Path(path_str)
            if path.is_relative_to(self.phantom_home):
                entry["user"] = PHANTOM_USER
                entry["group"] = PHANTOM_GROUP

        self.manifest.write(manifest)

    @classmethod
    def _run(
        cls,
        command,
        check=True,
        stdin=None,
        timeout: Optional[int] = 60,
        run_as_phantom_user=False,
        **kwargs,
    ):
        """Helper to run commands via subprocess.

        Not using InstallBase.run, because some of the code in this class depends on actual
        subprocess exceptions getting raised instead of InstallError.
        """
        if run_as_phantom_user:
            command = f"sudo -u {PHANTOM_USER} {command}"

        logger.debug(f"running command: {command}")
        proc_kwargs = {
            "shell": True,
            "encoding": "utf-8",
            "stdout": subprocess.PIPE,
            "stderr": subprocess.PIPE,
            "stdin": subprocess.PIPE,
        }
        proc_kwargs.update(kwargs)
        proc = subprocess.Popen(command, **proc_kwargs)  # noqa: PHANTOM112
        stdout, stderr = proc.communicate(stdin, timeout=timeout)
        proc.stdout = stdout
        proc.stderr = stderr

        logger.debug(stdout)

        if stderr:
            logger.debug(stderr)

        if proc.returncode != 0 and check:
            raise PhCalledProcessException(stderr, proc)

        return proc

    @classmethod
    def _run_sql(self, sql, db="", port=None, user="", run_as_phantom_user=False):
        """Run sql via psql.

        Can't use pg_utils.legacy_run_sql for all uses cases, because it doesn't really work when running as
        root on a partially/fully migrated system. It also doesn't check the exit codes on 4.9

        Before migration:
         - phantom user does not have access to postgres on port 6432, but has access to 'phantom' db
           via port 5432
         - postgres user has access on all dbs on port 5432 and to 'phantom' db on port 6432
        After migration:
         - phantom user has access to postgres and all dbs/tables on both ports
         - postgres user does not exist.

        So for read operations, it's pretty much always safe to use 'phantom' user via pgbouncer (port 6432),
        but alter operations need to consider the migration status at the point that the command runs.

        Also, db and role default to the name of the user running the command, so in most cases, it's
        not necessary to use those.
        """
        db = f"-d {db}" if db else ""

        if not run_as_phantom_user:
            su = "sudo -u postgres"
            port = port or 5432
        else:
            su = ""
            port = port or 6432

        if user:
            user = f"-U {user}"

        cmd = f"{su} psql -h /tmp -At -p {port} {user} {db}"

        return self._run(cmd, stdin=sql, run_as_phantom_user=run_as_phantom_user, timeout=10)

    def _is_dynamic_file(self, component: str, path: Path) -> bool:
        """Non-json files in the configuration package should have their paths expanded"""
        return path.is_file() and component == "configuration" and not path.name.endswith(".json")

    def _is_migration_entry(self, path: Path) -> bool:
        return path.is_relative_to(self.migration_component_dir)

    def _remove_directory(self, directory: Path) -> None:
        """Remove a directory from the filesystem and the manifest."""
        if not directory.is_dir():
            logger.debug(
                f"Attempted to remove a nonexistent directory '{str(directory)}'. Returning early."
            )
            return

        shutil.rmtree(directory, ignore_errors=True)
        self.manifest.remove_directory(directory)

    def _remove_priv_only_files(self):
        logger.info("Removing files that are only relevant for privileged installs.")
        entries_to_remove: List[Path] = []
        # Remove files on first pass
        for path_str, entry in self.manifest.read().items():
            path = Path(path_str)
            if (
                entry["priv_only"]
                and path.exists()
                and not path.is_dir()
                and not self._is_migration_entry(path)
            ):
                path.unlink()
                entries_to_remove.append(path)

        self.manifest.remove_entries(*entries_to_remove)
        entries_to_remove = []

        # Remove directories on second pass.
        for path_str, entry in reversed(list(self.manifest.read().items())):
            path = Path(path_str)
            if (
                entry["priv_only"]
                and not self._is_migration_entry(path)
                and path.is_dir()
                and not os.listdir(path)
            ):
                path.rmdir()
                entries_to_remove.append(path)

        self.manifest.remove_entries(*entries_to_remove)

    def install_files(self):

        self._remove_priv_only_files()
        logger.debug("Removing erlang")
        erlang_directory = Path(self.phantom_usr) / "erlang"
        self._remove_directory(erlang_directory)

        logger.info("Installing files required by unprivileged installs.")
        install_manifest = self.manifest.read()

        migration_entries: List[Path] = []
        migration_dir_entries: List[Path] = []
        new_entries: List[Path] = []

        for path_str, entry in install_manifest.items():
            current_path = Path(path_str)
            if self._is_migration_entry(current_path):
                try:
                    component = current_path.relative_to(self.migration_component_dir).parts[0]
                except IndexError:
                    continue

                component_dir = self.migration_component_dir / component
                if not component_dir.is_dir():
                    continue

                migration_entries.append(current_path)

                if component_dir == current_path:
                    migration_dir_entries.append(current_path)
                    continue

                relative_path = current_path.relative_to(component_dir)
                destination_path = self.phantom_home / relative_path

                if not current_path.exists() and not current_path.is_symlink():
                    if destination_path.exists():
                        continue
                    else:
                        raise Exception(
                            f"Unexpected error installing files. Neither the source path {current_path} or the destination path {destination_path} exists."
                        )

                if current_path.is_dir():
                    destination_path.mkdir(mode=entry["permissions"], parents=True, exist_ok=True)
                    migration_dir_entries.append(current_path)

                else:
                    destination_path.parent.mkdir(parents=True, exist_ok=True)
                    if self._is_dynamic_file(component, current_path):
                        content = self._expand_paths(current_path.read_text())
                        destination_path.write_text(content)
                        current_path.unlink()
                    else:
                        shutil.move(current_path, destination_path)

                perms_msg = ""
                # Do not chmod symlinks. Trust that the targets are in the manifest and will have their permissions
                # enforced.
                if not destination_path.is_symlink():
                    current_permissions = f"0o{oct(destination_path.stat().st_mode)[-3:]}"
                    destination_path.chmod(entry["permissions"])

                    if current_permissions != oct(entry["permissions"]):
                        perms_msg = f" Changed permissions from {current_permissions} to {oct(entry['permissions'])}"

                logger.debug(f"Processed '{destination_path}'. {perms_msg}")
                new_entries.append(destination_path)

        for dir_path in migration_dir_entries:
            if dir_path.exists():
                shutil.rmtree(dir_path, ignore_errors=True)

        self.manifest.remove_entries(*migration_entries)
        self.manifest.upsert_entries(*new_entries)

    def configure_nri(self):
        """Do necessary configuration for NRI.

        This is mostly installing/modifying configuration files, for example,
        updating logging paths from /var/log to <phantom home>/var/log"""
        logger.info(f'Configuring {PRODUCT_NAME} to run under the "{PHANTOM_USER}" user.')
        self._update_pgbouncer_confs()
        self._update_enable()
        self._move_ssh_keys()
        self._update_install_conf()
        self._update_dot_soar()
        self._update_proxyd()
        self._update_firewall()
        self._update_symlinks()
        self._ensure_full_name()
        self._change_file_ownerships()
        self._clean_up_manifest()
        self._install_erlang()
        self._set_system_limits()
        self._update_bashrc()
        if self.is_clustered:
            self._update_rabbit_and_consul_configs()
        self.start_phantom()
        self._update_crontab()  # Requires log dirs to be owned by phantom

    def configure_daemon_logging(self):
        """Configure daemons to log to files rather than to rsyslog.

        Note: This requires the database to be up!
        """
        logger.info("Configuring daemon logging.")

        daemon_settings = json.loads(
            self._run_sql(
                "select daemons from system_settings", run_as_phantom_user=True
            ).stdout.strip()
        )

        # Update log paths and log type
        for daemon, config in daemon_settings.items():
            config["log_path"] = str(self.phantom_log)
            config["log_type"] = "file"

        # Update workflowd rest port
        daemon_settings["workflowd"]["rest_post_port"] = self.external_https_port

        self._run_sql(
            f"update system_settings set daemons='{json.dumps(daemon_settings)}';",
            run_as_phantom_user=True,
        )

    def configure_postgres_roles(self):
        """Configure phantom-role to be a superuser role in the db.

        NRI expects 'phantom' role to be the superuser and to own all the databases/tables.

        Note: This only applies for local databases.
        """
        phantom_is_superuser = (
            self._run_sql(
                f"select 1 from pg_catalog.pg_user WHERE usename = '{PHANTOM_USER}' and usesuper=true;",
                run_as_phantom_user=True,
            ).stdout.strip()
            == "1"
        )

        if not phantom_is_superuser:
            logger.info("Configuring postgresql roles.")

            logger.debug('Transferring all db/table ownerships to "postgres" role.')
            current_roles = self._run_sql("SELECT usename FROM pg_catalog.pg_user").stdout.split()
            current_roles.remove("postgres")

            databases = self._run_sql(
                "select datname from pg_database where datistemplate=false"
            ).stdout.split()
            templates = self._run_sql(
                "select datname from pg_database where datistemplate=true"
            ).stdout.split()

            for template in templates + databases:
                self._run_sql(f"ALTER DATABASE {template} OWNER TO postgres")

            logger.debug('Removing extraneous database roles, including "phantom".')
            for role in current_roles:
                self._reassign_privs_and_drop_role(role, databases)

            logger.debug("Creating a temporary superuser in postgres.")

            tmp_superuser = "migration_superuser"
            pg_map = "migration_map"
            self._run_sql(
                f"CREATE ROLE {tmp_superuser} LOGIN SUPERUSER CREATEROLE CREATEDB REPLICATION BYPASSRLS"
            )

            # Enable phantom user to log in as tmp_superuser
            pg_ident = self.phantom_home / "data/db/pg_ident.conf"
            pg_ident_append = f"{pg_map}  {PHANTOM_USER}  {tmp_superuser}\n"

            pg_hba = self.phantom_home / "data/db/pg_hba.conf"
            pg_hba_append = (
                f"local  all  {tmp_superuser}  peer  map={pg_map}\n"
                f"local  all  {PHANTOM_USER}  peer\n"
            )

            with pg_ident.open("r+") as f:
                pg_ident_original = f.read()
                f.write(pg_ident_append)

            with pg_hba.open("a") as f:
                f.write(pg_hba_append)

            self._run(f"{self._phsvc} restart postgresql-11")

            logger.debug(f"Renaming postgres role to {PHANTOM_USER}")
            self._run_sql(
                f"ALTER ROLE postgres RENAME TO {PHANTOM_USER}",
                db="postgres",
                port=5432,
                user=tmp_superuser,
                run_as_phantom_user=True,
            )
            self._run_sql(f"DROP ROLE {tmp_superuser}", port=5432, run_as_phantom_user=True)

            # undo pg_ident modifications. pg_hba will be overwritten later in the migration.
            with pg_ident.open("w") as f:
                f.write(pg_ident_original)

    def _reassign_privs_and_drop_role(self, role, databases):
        # REASSIGN OWNED and DROP OWNED only affect the current database, so we need to execute
        # these on all databases.
        reassign_sql = f"REASSIGN OWNED BY {role} TO postgres"
        drop_owned_sql = f"DROP OWNED BY {role} CASCADE"

        for dbname in databases:
            self._run_sql(reassign_sql, db=dbname)
            self._run_sql(drop_owned_sql, db=dbname)

        self._run_sql(f"DROP ROLE {role}")

    def _install_erlang(self):
        erlang_location = self.phantom_usr / "erlang"
        self._run(
            f"{self._phenv} {erlang_location / 'Install'} -minimal {erlang_location}",
            run_as_phantom_user=True,
        )
        self.manifest.upsert_directory(erlang_location)

    def _update_bashrc(self):
        BEGIN_SECTION = (
            "##### BEGIN: SECTION MANAGED BY PHANTOM, DO NOT REMOVE ######################"
        )
        END_SECTION = (
            "##### END: SECTION MANAGED BY PHANTOM, DO NOT REMOVE ########################"
        )
        bash_rc = Path(PHANTOM_USER_HOME) / ".bashrc"
        source_enable_line = f"source {self.phantom_etc / 'enable'}"
        try:
            lines = bash_rc.read_text().strip().splitlines()
        except FileNotFoundError:
            lines = []

        try:
            start = lines.index(BEGIN_SECTION)
            end = lines.index(END_SECTION)
            section = lines[start + 1 : end]
            if source_enable_line not in section:
                lines = lines[0 : start + 1] + section + [source_enable_line] + lines[end:]
        except ValueError:
            lines += [BEGIN_SECTION, source_enable_line, END_SECTION]

        bash_rc.write_text("\n".join(lines) + "\n")

    def _update_install_conf(self):
        phantom_install_conf = PhantomInstallConf()
        phantom_install_conf.config.set(
            PhantomInstallConf.INI_SECTION_PHANTOM, PhantomInstallConf.INI_KEY_PLATFORM, "unpriv"
        )
        phantom_install_conf.config.set(
            PhantomInstallConf.INI_SECTION_PHANTOM,
            PhantomInstallConf.INI_KEY_SSL_PORT,
            self.local_https_port,
        )
        phantom_install_conf.config.set(
            PhantomInstallConf.INI_SECTION_PHANTOM,
            PhantomInstallConf.INI_KEY_REST_POST_PORT,
            self.local_https_port,
        )
        phantom_install_conf.write()

    def _update_dot_soar(self):
        """Update install info in .soar file."""
        dot_soar = self.phantom_home / ".soar"
        data = json.loads(dot_soar.read_text())
        data["platform"] = "unpriv"
        data["phantom_home"] = str(self.phantom_home)
        data["phantom_port"] = int(self.local_https_port)
        dot_soar.write_text(json.dumps(data))

    def _set_system_limits(self):
        limits_path = Path("/etc/security/limits.d/25-phantom-limits.conf")
        logger.debug(f"Creating limits file {limits_path}.")
        with open(limits_path, "w") as limits:
            limits.write(
                f"{PHANTOM_USER}   hard   nofile   64000\n"
                f"{PHANTOM_USER}   soft   nofile   64000\n"
                f"{PHANTOM_USER}   hard   nproc    64000\n"
                f"{PHANTOM_USER}   soft   nproc    64000\n"
            )

        logger.debug("Reload system settings.")
        self._run("sysctl --system")

    def _clean_up_manifest(self):
        logger.debug("Deleting files/folders that live outside PHANTOM_HOME from the manifest.")
        entries_to_remove: List[str] = []
        manifest = self.manifest.read()
        for path_str, entry in manifest.items():
            path = Path(path_str)
            if not path.is_relative_to(self.phantom_home):
                entries_to_remove.append(path_str)

        # delete without verifying that the files don't exist.
        for path in entries_to_remove:
            logger.debug(f"Deleting path from the manifest: {path}")
            del manifest[path]

        self.manifest.write(manifest)

    def _update_enable(self):
        """Update paths and misc vars in enable."""
        logger.debug("Updating enable.")
        conf = self.phantom_etc / "enable"
        for key, value in self._get_conf_expansions().items():
            self._modify_line_in_file(conf, rf"export {key}=.*", f"export {key}={value}")

    def _move_ssh_keys(self):
        """Move ssh keys from /var/cache/nginx/.ssh to /home/phantom/.ssh"""
        logger.debug("Moving nginx ssh keys.")
        source_dir = Path("/var/cache/nginx/.ssh")
        files = ["id_rsa", "id_rsa.pub"]
        target_dir = Path(PHANTOM_USER_HOME) / ".ssh"
        target_dir.mkdir(mode=0o700, exist_ok=True)

        for source_file in files:
            source_path = source_dir / source_file
            target_path = target_dir / source_file
            if not source_path.exists() and target_path.exists():
                logger.debug(f"Skipping {source_path} as it seems to have been moved already.")

            elif source_path.exists():
                if target_path.is_file():
                    target_path.unlink()

                shutil.move(source_path, target_path)
                shutil.chown(target_path, PHANTOM_USER, PHANTOM_USER)

            else:
                logger.error(f"Failed to find expected file: {source_path}.")

        self._run(
            f"chown -R {PHANTOM_USER}:{PHANTOM_GROUP} {target_dir}",
            timeout=None,
        )

    def _update_proxyd(self):
        logger.debug("Updating proxyd configuration.")
        proxy_daemon_sh = self.phantom_home / "proxy/bin/proxy_daemon.sh"
        self._modify_line_in_file(
            proxy_daemon_sh,
            r"^PROXY_PATH=.*",
            f'PROXY_PATH={self.phantom_home / "proxy"}',
        )
        self._modify_line_in_file(proxy_daemon_sh, r"^USER_ARG=.*", "USER_ARG=")

        proxy_conf = self.phantom_home / "proxy/local/proxy.conf"
        self._modify_line_in_file(
            proxy_conf,
            r"^product_management_port=.*",
            f"product_management_port={self.local_https_port}",
        )
        self._modify_line_in_file(
            proxy_conf, r"^proxy_daemon_log_dir=.*", f"proxy_daemon_log_dir={self.phantom_log}"
        )

    def _update_pgbouncer_confs(self):
        """Update pgbouncer config."""
        logger.debug("Updating PGBouncer configuration.")

        # For internal databases, change the user with which we access the database
        if self.external_db is False:
            logger.debug("Updating PGBouncer configuration.")
            pgbouncer_database_ini = (
                self.phantom_home / "etc" / "pgbouncer" / "pgbouncer.database.ini"
            )
            db_settings = {
                "phantom": f"user={PHANTOM_USER}",
                "postgres": f"user={PHANTOM_USER}",
            }
            config = RawConfigParser(allow_no_value=True)
            config.read(pgbouncer_database_ini)
            for db_name, db_config in db_settings.items():
                config.set("databases", db_name, db_config)

            with open(pgbouncer_database_ini, "w") as f:
                config.write(f)

            pgbouncer_database_ini.chmod(0o750)

    def _update_symlinks(self):
        """Update python/python3/pip/pip3/phantomjs symlinks in usr/bin.

        This is required, because tar -xzf removes any files that it would otherwise
        overwrite, so when we extract the dependencies over the current files, the symlinks
        lose their targets and stop working.
        """
        logger.debug("Updating symlinks in <PHANTOM_HOME>/usr/bin.")
        links_to_create = {
            "usr/bin/python3": "usr/python39/bin/python3.9",
            "usr/bin/python": "usr/python39/bin/python3.9",
            "usr/bin/pip3": "usr/python39/bin/pip3",
            "usr/bin/pip": "usr/python39/bin/pip3",
            "bin/2to3": "usr/python39/bin/2to3",
        }
        entries_to_update = []
        for link_str, path_str in links_to_create.items():
            source = self.phantom_home / path_str
            link = self.phantom_home / link_str
            if not link.parent.exists():
                logger.debug(f"Creating directory {link.parent}")
                link.parent.mkdir(parents=True)

            try:
                # link.exists() returns False if a symlink exists but points to nonexistent path
                link.unlink()
                logger.debug(f"Removed existing symlink: {link}")
            except FileNotFoundError:
                pass

            logger.debug(f"Creating symlink: {link} -> {source}")
            link.symlink_to(source)
            entries_to_update.append(link)

        self.manifest.upsert_entries(*entries_to_update)

        links_to_delete = [
            Path("/usr/bin/phantomjs"),
            Path("/bin/phenv"),
            Path("/bin/jq"),
            Path("/vault"),
        ]
        logger.debug("Removing symlinks added by the privileged installer.")
        for link in links_to_delete:
            link.unlink(missing_ok=True)

        self.manifest.remove_entries(*links_to_delete)

    def _ensure_full_name(self):
        """Ensure that phantom-user has a gecos/fullname set.

        This is required for git to work. On Privileged systems uwsgi runs under 'nginx' user,
        which has this set, but on a migrated system, uwsgi runs under <phantom user>, which by default
        does not have this attribute set.
        """
        if not pwd.getpwnam(PHANTOM_USER).pw_gecos:
            self._run(f'chfn -f "Phantom user" {PHANTOM_USER}')  # noqa: PHANTOM110

    def _update_firewall(self):
        """Update firewall to allow access on the chosen port.

        Also forwards 443 to this port, so people can continue to use the port they've used before.
        """
        firewalld_not_installed = self._run("command -v firewall-cmd", check=False).returncode != 0
        if firewalld_not_installed:
            logger.warning(
                f"Skipping firewall setup due to firewalld not being installed. SOAR will not be listening on port 443 and will only be available on port {self.local_https_port}")
            return
        logger.debug(f"Opening port {self.local_https_port} in firewalld and forwarding 443 to it.")
        # These commands are idempotent.
        self._run(f"firewall-cmd --permanent --zone=public --add-port={self.local_https_port}/tcp")
        self._run(
            f"firewall-cmd --permanent --zone=public --add-forward-port=port=443:proto=tcp:toport={self.local_https_port}"
        )
        self._run("systemctl reload firewalld")

    def remove_migration_component(self):
        self._remove_directory(self.migration_component_dir)

    def _expand_paths(self, content: str) -> str:
        """
        Replace our variable placeholder strings in files with their actual values
        """
        for name, value in self._get_conf_expansions().items():
            content = content.replace(f"__{name}__", str(value))
        return content

    def _get_conf_expansions(self) -> dict:
        return {
            "PHANTOM_DB_NAME": "phantom",
            "PHANTOM_DB_SCHEMA": "public",
            "PHANTOM_FIPS_MODE": "0",
            "PHANTOM_GROUP": PHANTOM_GROUP,
            "PHANTOM_HOME": PHANTOM_HOME,
            "PHANTOM_USER": PHANTOM_USER,
            "PHANTOM_WEBSERVER_USER": PHANTOM_USER,
            "PHANTOM_WEBSERVER_GROUP": PHANTOM_GROUP,
            "PHANTOM_IS_CLOUD": 0,
            "PHANTOM_IS_NONROOT_INSTALL": 1,
            "PHANTOM_USER_HOME": PHANTOM_USER_HOME,
            "PHANTOM_VAR": self.phantom_var,
            "PHANTOM_ETC": self.phantom_etc,
            "PHANTOM_USR": self.phantom_usr,
            "PHANTOM_HTTPS_PORT": self.local_https_port,
        }

    def manage_py(self, cmd: str):
        cmd = f"{self._phenv} python3 -m manage {cmd}"

        return self._run(cmd, run_as_phantom_user=True)

    def _update_crontab(self):
        """Update crontab for phantom-user.

        Most of the crontab entries are the same on priv vs NRI.
        We just need to add the "start phantom on boot" entry and the logrotate entry.
        """
        logger.debug("Updating crontab.")
        CronJob = namedtuple("CronJob", ["name", "command"])
        ScheduledTask = namedtuple("ScheduledTask", ["name", "management_cmd", "cron_str"])
        DataRetentionEntry = namedtuple(
            "DataRetentionEntry", ["model", "frequency", "unit", "error_msg"]
        )

        proc = self._run("crontab -r", check=False, run_as_phantom_user=True)

        # Job definitions
        data_retention_entries = [
            DataRetentionEntry(
                "device_profile", 1, "hours", "Failed to schedule Mobile Registration Cleanup"
            )
        ]
        scheduled_tasks = [
            ScheduledTask("db_maintenance", "db_maintenance", "0 23 * * 0"),
        ]

        log_rotate_command = (
            f"* * * * * /usr/sbin/logrotate "
            f'-s {self.phantom_home / "private/logrotate.status"} '
            f'{self.phantom_home / "etc/logrotate.d/phantom_logrotate.conf"}'
        )
        scheduler_command = self.manage_py("phantom_scheduler --cron").stdout.strip()
        cron_jobs = [
            CronJob("logrotate", log_rotate_command),
            CronJob(
                "start_phantom",
                f'@reboot {self.phantom_home / "bin/start_phantom.sh"}',
            ),
            CronJob("phantom_scheduler", scheduler_command),
        ]

        # Actual cron creation steps
        for task in scheduled_tasks:
            command = (
                f'{self.phantom_home / "bin/phenv"} python '
                f'{self.phantom_home / "www/manage.py"} {task.management_cmd}'
            )
            self.manage_py(
                f"phantom_scheduler --add "
                f'--name "{task.name}" '
                f'--command "{command}" '
                f'--cron-schedule "{task.cron_str}"',
            )

        cron_proc = self._run("crontab -l", run_as_phantom_user=True, check=False)
        if proc.stderr:
            crontab = []
        else:
            crontab = cron_proc.stdout.split("\n")

        crontab.append(CRON_HEADER)
        for cron_job in cron_jobs:
            if cron_job.command not in crontab:
                logger.debug(f"Adding job to crontab: {cron_job.name}")
                crontab.append(cron_job.command)
            else:
                logger.debug(f"Job already in crontab: {cron_job.name}")
        crontab.append(CRON_FOOTER)

        # Finally, handle every data retention job
        for entry in data_retention_entries:
            self.manage_py(
                f"configure_db_maintenance --add -m {entry.model} -u {entry.unit} -a {entry.frequency}",
            )

        crontab_text = "\n".join([entry for entry in crontab if entry]).strip() + "\n"
        self._run("crontab -", stdin=crontab_text, run_as_phantom_user=True)

        # Copy crontab to a location under Phantom home -- useful for reprovisions
        crontab_path = self.phantom_home / "etc/phantom.cron"
        crontab_path.write_text(crontab_text)
        self.manifest.upsert_entries(crontab_path)

    def _update_rabbit_and_consul_configs(self):
        shutil.move(
            cluster_constants.SUPERVISOR_CONSUL_PATH,
            Path(cluster_constants.SUPERVISOR_TARGET_PATH)
            / cluster_constants.SUPERVISOR_CONSUL_CONF,
        )
        shutil.move(
            cluster_constants.SUPERVISOR_RABBIT_PATH,
            Path(cluster_constants.SUPERVISOR_TARGET_PATH)
            / cluster_constants.SUPERVISOR_RABBIT_CONF,
        )
        # need to update the env-file too.
        rabbitmq_env_conf = self.phantom_usr / "rabbitmq/etc/rabbitmq/rabbitmq-env.conf"
        self._modify_line_in_file(rabbitmq_env_conf, r'"/var', f'"{self.phantom_var}')


def _get_nri_nginx_listen_port():
    try:
        nginx_conf = Path("/opt/phantom/usr/nginx/conf/conf.d/phantom-nginx-server.conf")
        with nginx_conf.open("r") as f:
            nginx_conf_content = f.read()
        listen_line = re.search(r"listen.*ssl.*;", nginx_conf_content).group()
        return int(re.search(r"\d+", listen_line).group())

    except (FileNotFoundError, AttributeError) as e:
        return None


def _validate_https_port(value):
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("Port number must be an integer.")

    if value < 1024 or value > 65535:
        raise argparse.ArgumentTypeError(
            "Port number must be in range 1024-65535, so it can be used without root privileges."
        )

    current_nginx_port = _get_nri_nginx_listen_port()
    if current_nginx_port and value != current_nginx_port:
        # Check that the port is free
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.bind(("127.0.0.1", value))
            sock.listen(5)
            sock.close()
        except socket.error:
            raise argparse.ArgumentTypeError(
                "Port number already in use. Please shutdown the service using that port or use another port."
            )

    return value


if __name__ == "__main__":
    # Log to current folder, since we'll be moving the log folder with this script.
    filename = (
        f"privileged_to_nri_migration_{datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')}.log"
    )
    filepath = Path(PHANTOM_HOME) / filename

    init_logging(
        __name__,
        console_log_config={"formatter": __name__},
        file_log_config={"filename": filepath, "level": logging.DEBUG, "add_to_root": True},
        formatter={"format": "{message}", "style": "{"},
        set_as_default=False,
        root_logger_level="DEBUG",
    )

    description = (
        f"This script will migrate your privileged {PRODUCT_NAME} "
        f"installation to an unprivileged one.\n{MIGRATION_DESCRIPTION}"
    )

    argparser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    argparser.add_argument(
        "-p",
        "--https-port",
        help="Https port that nginx should bind to.",
        type=_validate_https_port,
        default=8443,
    )
    argparser.add_argument(
        "-y",
        "--no-prompt",
        action="store_true",
        default=False,
        help="Do not prompt for confirmation before migrating.",
    )
    argparser.add_argument(
        "--cluster-phase",
        default="",
        type=str,
        help=argparse.SUPPRESS,
    )
    args = argparser.parse_args()

    if os.geteuid() != 0:
        logger.error("Script must be run as root. Exiting.")
        sys.exit(-1)

    try:
        universal_migrator = MigratePrivilegedInstallToUnprivileged(**vars(args))
        universal_migrator.migrate()

    except UserDeclinedToContinue:
        sys.exit(-1)

    except Exception as e:
        logger.debug(str(e), exc_info=True)
        logger.error(
            f"Migration failed due to an unexpected error ({str(e)}).\nYou can find the full log here: \n{filepath}\n"
        )
        sys.exit(-1)

    else:
        target = Path("/opt/phantom/var/log/phantom") / filename
        logger.info(f"\nMoving migration log to {target}")
        shutil.move(filepath, "/opt/phantom/var/log/phantom")
